Function INDMED (XDONT)
!  Returns index of median value of XDONT.
!  Finds the index of the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 10% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
      Real, Dimension (:), Intent (In) :: XDONT
      Integer :: INDMED
! __________________________________________________________
      Real (Kind(XDONT)), Parameter :: XHUGE = HUGE (XDONT)
      Integer, Dimension (7*((SIZE(XDONT)+6)/7)) :: INDST
      Integer, Dimension (:), Allocatable :: ISTRT, IENDT, IWRKT
      Integer :: NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, IDCR, ICRS, ICRS1, ICRS2, IMAX, IMIN
      Integer :: IWRK, IWRK1, NWRK, IMED1, IMED7, IDONA, IDONB
      Integer :: IWRK2, IWRK3, IWRK4, IWRK5, IWRK6, IWRKU, IWRKL
      Real (Kind(XDONT)) :: XWRK, XWRK1, XWRK2, XWRK3, XWRK4, XWRK5, XWRK6
      Real (Kind(XDONT)) :: XMED7, XMAX, XMIN, XWRKL, XWRKU
      Integer :: NDON
!
      NDON = SIZE (XDONT)
      If (NDON < 35) Then
!
         Do IDON = 1, SIZE(XDONT)
            INDST (IDON) = IDON
         End Do
         INDMED = i_med (INDST, SIZE(XDONT))
      Else
         NMED = (NDON+1) / 2
!
         If ((7*(NDON/7)) < NDON) Then
            NTRI = 7*(NDON/7) + 7
         Else
            NTRI = NDON
         Endif
         NWRK = (NTRI+6)/7
         NWRK = 7*((NWRK+6)/7)
         Allocate (ISTRT (1:NWRK))
         Allocate (IENDT (1:NWRK))
         Allocate (IWRKT (1:NWRK))
!
!  Make sorted subsets of 7 elements
! This routine uses a sort of Shellsort in a first pass
! trying to bring the set to the following form:
! {minimum} {X1 . . } {X4 . .} with X1 <= X4
! then
! {minimum} {X1 X2 . } {X4 X5 .} with X2 <= X5
! then sort each of the subsets, and then
! {minimum} {second minimum X2 X3 } {X4 X5 X6=maximum} with X2 <= X3 and
!                                                           X4 <= X5
!  and lastly merge the two subsets.
!
         IMAX = 1
         IMIN = 1
         XMAX = XDONT (IMAX)
         XMIN = XDONT (IMIN)
         IDEB = 1
         IDON1 = NDON/7
         DO ICRS = 1, IDON1
            IWRK = IDEB
            XWRK = XDONT (IWRK)
            IDONA = IDEB+1
            IDONB = IDONA+3
            If (XDONT (IDONA) < XDONT (IDONB)) Then
              IWRKL = IDONA
              IWRKU = IDONB
              XWRKL = XDONT (IWRKL)
              XWRKU = XDONT (IWRKU)
            Else
              IWRKU = IDONA
              IWRKL = IDONB
              XWRKU = XDONT (IWRKU)
              XWRKL = XDONT (IWRKL)
            End If
            If (XWRKL < XWRK ) Then
               If (XWRKU < XWRK) Then
                    IWRK1 = IWRKU
                    IWRK4 = IWRK
                    XWRK1 = XWRKU
                    XWRK4 = XWRK
                 Else
                    IWRK1 = IWRK
                    IWRK4 = IWRKU
                    XWRK1 = XWRK
                    XWRK4 = XWRKU
                 Endif
                 IWRK = IWRKL
                 XWRK = XWRKL
            Else
                 IWRK1 = IWRKL
                 IWRK4 = IWRKU
                 XWRK1 = XWRKL
                 XWRK4 = XWRKU
            End If
!
            IDONA = IDEB+2
            IDONB = IDONA+3
            If (XDONT (IDONA) < XDONT (IDONB)) Then
              IWRKL = IDONA
              IWRKU = IDONB
              XWRKL = XDONT (IDONA)
              XWRKU = XDONT (IDONB)
            Else
              IWRKU = IDONA
              IWRKL = IDONB
              XWRKU = XDONT (IDONA)
              XWRKL = XDONT (IDONB)
            End If
            If (XWRKL < XWRK ) Then
                IWRK2 = IWRK1
                XWRK2 = XWRK1
                If (XWRKU < XWRK) Then
                 IWRK1 = IWRKU
                 IWRKU = IWRK
                 XWRK1 = XWRKU
                 XWRKU = XWRK
                Else
                 IWRK1 = IWRK
                 XWRK1 = XWRK
                Endif
                IWRK = IWRKL
                XWRK = XWRKL
              Else
                If (XWRKL >= XWRK1) Then
                   IWRK2 = IWRKL
                   XWRK2 = XWRKL
                Else
                   IWRK2 = IWRK1
                   IWRK1 = IWRKL
                   XWRK2 = XWRK1
                   XWRK1 = XWRKL
                Endif
              End If
              If (XWRKU >= XWRK4 ) Then
                IWRK5 = IWRKU
                XWRK5 = XWRKU
              Else
                IWRK5 = IWRK4
                IWRK4 = IWRKU
                XWRK5 = XWRK4
                XWRK4 = XWRKU
              End If
!
              IDONA = IDEB+3
              IDONB = IDONA+3
              If (XDONT (IDONA) < XDONT (IDONB)) Then
                IWRKL = IDONA
                IWRKU = IDONB
                XWRKL = XDONT (IDONA)
                XWRKU = XDONT (IDONB)
              Else
                IWRKU = IDONA
                IWRKL = IDONB
                XWRKU = XDONT (IDONA)
                XWRKL = XDONT (IDONB)
              End If
              If (XWRKL <= XWRK ) Then
                IWRK3 = IWRK2
                IWRK2 = IWRK1
                XWRK3 = XWRK2
                XWRK2 = XWRK1
                If (XWRKU < XWRK) Then
                   IWRK1 = IWRKU
                   IWRKU = IWRK
                   XWRK1 = XWRKU
                   XWRKU = XWRK
                Else
                   IWRK1 = IWRK
                   XWRK1 = XWRK
                Endif
                IWRK = IWRKL
                XWRK = XWRKL
              Else
                If (XWRKL >= XWRK2) Then
                   IWRK3 = IWRKL
                   XWRK3 = XWRKL
                Else
                   IWRK3 = IWRK2
                   XWRK3 = XWRK2
                   If (XWRKL >= XWRK1) Then
                      IWRK2 = IWRKL
                      XWRK2 = XWRKL
                   Else
                      IWRK2 = IWRK1
                      IWRK1 = IWRKL
                      XWRK2 = XWRK1
                      XWRK1 = XWRKL
                   Endif
                Endif
              End If
              If (XWRKU >= XWRK5 ) Then
                IWRK6 = IWRKU
                XWRK6 = XWRKU
              Else
                IWRK6 = IWRK5
                XWRK6 = XWRK5
                If (XWRKU >= XWRK4) Then
                   IWRK5 = IWRKU
                   XWRK5 = XWRKU
                Else
                   IWRK5 = IWRK4
                   IWRK4 = IWRKU
                   XWRK5 = XWRK4
                   XWRK4 = XWRKU
                Endif
              End If
              If (XWRK < XMIN) Then
                  IMIN = IWRK
                  XMIN = XWRK
              End If
              INDST (IDEB) = IWRK
              INDST (IDEB+1) = IWRK1
              If (XWRK4 >= XWRK3) Then
                 INDST (IDEB+2) = IWRK2
                 INDST (IDEB+3) = IWRK3
                 INDST (IDEB+4) = IWRK4
                 INDST (IDEB+5) = IWRK5
              Else
                 If (XWRK4 >= XWRK2) Then
                    INDST (IDEB+2) = IWRK2
                    INDST (IDEB+3) = IWRK4
                 Else
                    INDST (IDEB+2) = IWRK4
                    INDST (IDEB+3) = IWRK2
                 Endif
                 If (XWRK3 <= XWRK5) Then
                    INDST (IDEB+4) = IWRK3
                    INDST (IDEB+5) = IWRK5
                 Else
                    INDST (IDEB+4) = IWRK5
                    INDST (IDEB+5) = IWRK3
                 Endif
              Endif
              INDST (IDEB+6) = IWRK6
              If (XWRK6 > XMAX) Then
                  IMAX = IWRK6
                  XMAX = XWRK6
              End If
              IWRKT(ICRS) = INDST (IDEB + 3)
              IDEB = IDEB + 7
         End Do
!
!  Add-up alternatively MAX and MIN values to make the number of data
!  an exact multiple of 7. 
!
         IDEB = 7 * (NDON/7)
         If (IDEB < NDON) Then
!
            Do ICRS = IDEB+1, NDON
               XWRK1 = XDONT (ICRS)
               IF (XWRK1 > XMAX) Then
                  IMAX = ICRS
                  XMAX = XWRK1
               End If
               IF (XWRK1 < XMIN) Then
                  IMIN = ICRS
                  XMIN = XWRK1
               End If
               INDST (ICRS) = ICRS
            End Do
            IWRK1 = IMAX
            Do ICRS = NDON+1, IDEB+7
               INDST (ICRS) = IWRK1
               If (IWRK1 == IMAX) Then
                  IWRK1 = IMIN
               Else
                  IWRK1 = IMAX
               End If
            End Do
!
            Do ICRS = IDEB+2, IDEB+7
               IWRK = INDST (ICRS)
               XWRK = XDONT (IWRK)
               Do IDCR = ICRS - 1, IDEB+1, - 1
                  If (XWRK >= XDONT (INDST(IDCR))) Exit
                  INDST (IDCR+1) = INDST (IDCR)
               End Do
               INDST (IDCR+1) = IWRK
            End Do
            IDON1 = IDON1 + 1
            IWRKT (IDON1) = INDST (IDEB + 3)
!
         End If
!
!  Find XMED7, the median of the medians
!
         IMED7 = i_med (IWRKT, IDON1)
         XMED7 = XDONT (IMED7)
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         NMED = (NTRI+1) / 2
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            IWRK3 = INDST (IDON+3)
            If (XDONT (IWRK3) > XMED7) Then
                  IMED = IMED - 2
                  If (XDONT (INDST (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDONT (INDST (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XDONT (IWRK3) < XMED7) Then
                  IMED = IMED + 2
                  If (XDONT (INDST (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDONT (INDST (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            XWRK3 = XDONT (INDST (IMED))
            If (XWRK3 > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XWRK3 < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XDONT (INDST (IMED)) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XDONT (INDST (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XDONT (INDST (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
         NMED = (NTRI+1) / 2
         If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                IWRK1 = IMAX
                XWRK1 = XDONT (IWRK1)
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
                Do IDON = 1, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDONT (INDST (ICRS)) < XWRK1) Then
                            IWRK = INDST (ICRS)
                            XWRK = XDONT (IWRK)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDONT (INDST (IDCR))) Exit
                               INDST  (IDCR+1) = INDST (IDCR)
                            End Do
                            INDST (IDCR+1) = IWRK
                            IWRK1 = INDST (ICRS1)
                            XWRK1 = XDONT (IWRK1)
                         Else
                           If (ICRS2 < NORD) Then
                              INDST (ICRS1) = INDST (ICRS)
                              IWRK1 = INDST (ICRS1)
                              XWRK1 = XDONT (IWRK1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                      If (ICRS2 >= NORD) Exit
                End Do
                Do IDON = IDON+7, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDONT (INDST (ICRS)) >= XWRK1) Exit
                         IWRK = INDST (ICRS)
                         XWRK = XDONT (IWRK)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDONT (INDST (IDCR))) Exit
                               INDST  (IDCR+1) = INDST (IDCR)
                         End Do
                         INDST (IDCR+1) = IWRK
                         IWRK1 = INDST (ICRS1)
                         XWRK1 = XDONT (IWRK1)
                      End Do
                End Do
                INDMED = IWRK1
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
            Else
                INDMED = IMED7
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XDONT(INDST (ICRS)) > XWRK1) Then
                            IWRK = INDST (ICRS)
                            XWRK = XDONT (IWRK)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDONT(INDST (IDCR))) Exit
                               INDST (IDCR+1) = INDST (IDCR)
                            End Do
                            INDST (IDCR+1) = IWRK
                            IWRK1 = INDST(ICRS1)
                            XWRK1 = XDONT(IWRK1)
                         Else
                            If (ICRS2 < NORD) Then
                               INDST (ICRS1) = INDST (ICRS)
                               IWRK1 = INDST(ICRS1)
                               XWRK1 = XDONT(IWRK1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                      If (ICRS2 >= NORD) Exit
                End Do
                Do IDON = IDON+7, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XDONT(INDST (ICRS)) <= XWRK1) Exit
                         IWRK = INDST (ICRS)
                         XWRK = XDONT (IWRK)
                         IDCR = ICRS1 - 1
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDONT(INDST (IDCR))) Exit
                               INDST (IDCR+1) = INDST (IDCR)
                         End Do
                         INDST (IDCR+1) = IWRK
                         IWRK1 = INDST(ICRS1)
                         XWRK1 = XDONT(IWRK1)
                      End Do
                End Do
!
                INDMED = IWRK1
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
            End If
      End If
!
Contains
   Recursive Function i_med (IDONT, NDON) Result (ires_med)
!  Finds the index of the median of XDONT using the recursive procedure
!  described in Knuth, The Art of Computer Programming,
!  vol. 3, 5.3.3 - This procedure is linear in time, and
!  does not require to be able to interpolate in the
!  set as the one used in INDNTH. It also has better worst
!  case behavior than INDNTH, but is about 10% slower in
!  average for random uniformly distributed values.
! __________________________________________________________
      Integer, Dimension (:), Intent (InOut) :: IDONT
      Integer, Intent (In) :: NDON
      Integer :: ires_med
! __________________________________________________________
      Integer, Dimension (:), Allocatable :: ISTRT, IENDT, IWRKT
      Integer :: NTRI, NMED, NORD, NEQU, NLEQ, IMED, IDON, IDON1
      Integer :: IDEB, ITMP, IDCR, ICRS, ICRS1, ICRS2, IMAX, IMIN
      Integer :: IWRK, IWRK1, NWRK, IMED1, IMED7
      Real (Kind(XDONT)) :: XWRK, XWRK1, XMED7, XMAX, XMIN
!
      NMED = (NDON+1) / 2
!
!  If the number of values is small, then use insertion sort
!
      If (NDON < 35) Then
!
!  Bring minimum to first location to save test in decreasing loop
!
         IDCR = NDON
         If (XDONT (IDONT (1)) < XDONT (IDONT (IDCR))) Then
            IWRK = IDONT (1)
         Else
            IWRK = IDONT (IDCR)
            IDONT (IDCR) = IDONT (1)
         Endif
         XWRK = XDONT (IWRK)
         Do ITMP = 1, NDON - 2
            IDCR = IDCR - 1
            IWRK1 = IDONT (IDCR)
            XWRK1 = XDONT (IWRK1)
            If (XWRK1 < XWRK) Then
                IDONT (IDCR) = IWRK
                XWRK = XWRK1
                IWRK = IWRK1
            Endif
         End Do
         IDONT (1) = IWRK
!
! Sort the first half, until we have NMED sorted values
!
         Do ICRS = 3, NMED
            XWRK = XDONT (IDONT (ICRS))
            IWRK = IDONT (ICRS)
            IDCR = ICRS - 1
            Do 
                  If (XWRK >= XDONT (IDONT(IDCR))) Exit
                  IDONT (IDCR+1) = IDONT (IDCR)
                  IDCR = IDCR - 1
            End Do
            IDONT (IDCR+1) = IWRK
         End Do
!
!  Insert any value less than the current median in the first half
!
         XWRK1 = XDONT (IDONT (NMED))
         Do ICRS = NMED+1, NDON
            XWRK = XDONT (IDONT (ICRS))
            IWRK = IDONT (ICRS)
            If (XWRK < XWRK1) Then
               IDCR = NMED - 1
               Do 
                  If (XWRK >= XDONT (IDONT(IDCR))) Exit
                  IDONT (IDCR+1) = IDONT (IDCR)
                  IDCR = IDCR - 1
               End Do
               IDONT (IDCR+1) = IWRK
               XWRK1 = XDONT (IDONT (NMED))
            End If
         End Do
         ires_med = IDONT (NMED)
         Return
      End If
!
      If ((7*(NDON/7)) < NDON) Then
         NTRI = 7*(NDON/7) + 7
      Else
         NTRI = NDON
      Endif
      NWRK = (NTRI+6)/7
      NWRK = 7*((NWRK+6)/7)
      Allocate (ISTRT (1:NWRK))
      Allocate (IENDT (1:NWRK))
      Allocate (IWRKT (1:NWRK))
!
!  Make sorted subsets of 7 elements
! This routine uses a sort of Shellsort in a first pass
! trying to bring the set to the following form:
! {minimum} {X1 . . } {X4 . .} with X1 <= X4
! then
! {minimum} {X1 X2 . } {X4 X5 .} with X2 <= X5
! then sort each of the subsets, and then
! {minimum} {second minimum X2 X3 } {X4 X5 X6=maximum} with X2 <= X3 and
!                                                           X4 <= X5
!  and lastly merge the two subsets.
!
      IMAX = 1
      IMIN = 1
      XMAX = XDONT (IDONT(IMAX))
      XMIN = XDONT (IDONT(IMIN))
      IDEB = 1
      IDON1 = NDON/7
      DO ICRS = 1, IDON1
         IWRK = IDONT (IDEB)
         XWRK = XDONT (IWRK)
         IDONA = IDONT (IDEB+1)
         IDONB = IDONT (IDEB+4)
         If (XDONT (IDONA) < XDONT (IDONB)) Then
           IWRKL = IDONA
           IWRKU = IDONB
           XWRKL = XDONT (IWRKL)
           XWRKU = XDONT (IWRKU)
         Else
           IWRKU = IDONA
           IWRKL = IDONB
           XWRKU = XDONT (IWRKU)
           XWRKL = XDONT (IWRKL)
         End If
         If (XWRKL < XWRK ) Then
            If (XWRKU < XWRK) Then
                 IWRK1 = IWRKU
                 IWRK4 = IWRK
                 XWRK1 = XWRKU
                 XWRK4 = XWRK
              Else
                 IWRK1 = IWRK
                 IWRK4 = IWRKU
                 XWRK1 = XWRK
                 XWRK4 = XWRKU
              Endif
              IWRK = IWRKL
              XWRK = XWRKL
         Else
              IWRK1 = IWRKL
              IWRK4 = IWRKU
              XWRK1 = XWRKL
              XWRK4 = XWRKU
         End If
         IDONA = IDONT (IDEB+2)
         IDONB = IDONT (IDEB+5)
         If (XDONT (IDONA) < XDONT (IDONB)) Then
           IWRKL = IDONA
           IWRKU = IDONB
           XWRKL = XDONT (IDONA)
           XWRKU = XDONT (IDONB)
         Else
           IWRKU = IDONA
           IWRKL = IDONB
           XWRKU = XDONT (IDONA)
           XWRKL = XDONT (IDONB)
         End If
         If (XWRKL < XWRK ) Then
                IWRK2 = IWRK1
                XWRK2 = XWRK1
                If (XWRKU < XWRK) Then
                 IWRK1 = IWRKU
                 IWRKU = IWRK
                 XWRK1 = XWRKU
                 XWRKU = XWRK
                Else
                 IWRK1 = IWRK
                 XWRK1 = XWRK
                Endif
                IWRK = IWRKL
                XWRK = XWRKL
           Else
                If (XWRKL >= XWRK1) Then
                   IWRK2 = IWRKL
                   XWRK2 = XWRKL
                Else
                   IWRK2 = IWRK1
                   IWRK1 = IWRKL
                   XWRK2 = XWRK1
                   XWRK1 = XWRKL
                Endif
           End If
           If (XWRKU >= XWRK4 ) Then
                IWRK5 = IWRKU
                XWRK5 = XWRKU
           Else
                IWRK5 = IWRK4
                IWRK4 = IWRKU
                XWRK5 = XWRK4
                XWRK4 = XWRKU
           End If
         IDONA = IDONT (IDEB+3)
         IDONB = IDONT (IDEB+6)
         If (XDONT (IDONA) < XDONT (IDONB)) Then
           IWRKL = IDONA
           IWRKU = IDONB
           XWRKL = XDONT (IDONA)
           XWRKU = XDONT (IDONB)
         Else
           IWRKU = IDONA
           IWRKL = IDONB
           XWRKU = XDONT (IDONA)
           XWRKL = XDONT (IDONB)
         End If
         If (XWRKL <= XWRK ) Then
                IWRK3 = IWRK2
                IWRK2 = IWRK1
                XWRK3 = XWRK2
                XWRK2 = XWRK1
                If (XWRKU < XWRK) Then
                   IWRK1 = IWRKU
                   IWRKU = IWRK
                   XWRK1 = XWRKU
                   XWRKU = XWRK
                Else
                   IWRK1 = IWRK
                   XWRK1 = XWRK
                Endif
                IWRK = IWRKL
                XWRK = XWRKL
           Else
                If (XWRKL >= XWRK2) Then
                   IWRK3 = IWRKL
                   XWRK3 = XWRKL
                Else
                   IWRK3 = IWRK2
                   XWRK3 = XWRK2
                   If (XWRKL >= XWRK1) Then
                      IWRK2 = IWRKL
                      XWRK2 = XWRKL
                   Else
                      IWRK2 = IWRK1
                      IWRK1 = IWRKL
                      XWRK2 = XWRK1
                      XWRK1 = XWRKL
                   Endif
                Endif
           End If
           If (XWRKU >= XWRK5 ) Then
                IWRK6 = IWRKU
                XWRK6 = XWRKU
           Else
                IWRK6 = IWRK5
                XWRK6 = XWRK5
                If (XWRKU >= XWRK4) Then
                   IWRK5 = IWRKU
                   XWRK5 = XWRKU
                Else
                   IWRK5 = IWRK4
                   IWRK4 = IWRKU
                   XWRK5 = XWRK4
                   XWRK4 = XWRKU
                Endif
           End If
           If (XWRK < XMIN) Then
             IMIN = IWRK
             XMIN = XWRK
           End If
           IDONT (IDEB) = IWRK
           IDONT (IDEB+1) = IWRK1
           If (XWRK4 >= XWRK3) Then
              IDONT (IDEB+2) = IWRK2
              IDONT (IDEB+3) = IWRK3
              IDONT (IDEB+4) = IWRK4
              IDONT (IDEB+5) = IWRK5
           Else
              If (XWRK4 >= XWRK2) Then
                 IDONT (IDEB+2) = IWRK2
                 IDONT (IDEB+3) = IWRK4
              Else
                 IDONT (IDEB+2) = IWRK4
                 IDONT (IDEB+3) = IWRK2
              Endif
              If (XWRK3 <= XWRK5) Then
                 IDONT (IDEB+4) = IWRK3
                 IDONT (IDEB+5) = IWRK5
              Else
                 IDONT (IDEB+4) = IWRK5
                 IDONT (IDEB+5) = IWRK3
              Endif
           Endif
           IDONT (IDEB+6) = IWRK6
           If (XWRK6 > XMAX) Then
             IMAX = IWRK6
             XMAX = XWRK6
           End If
           IWRKT(ICRS) = IDONT (IDEB + 3)
           IDEB = IDEB + 7
      End Do
!
!  Add-up alternatively MAX and MIN values to make the number of data
!  an exact multiple of 7. 
!
      IDEB = 7 * (NDON/7)
      If (IDEB < NDON) Then
!
         Do ICRS = IDEB+1, NDON
            XWRK1 = XDONT (IDONT (ICRS))
            IF (XWRK1 > XMAX) Then
               IMAX = IDONT (ICRS)
               XMAX = XWRK1
            End If
            IF (XWRK1 < XMIN) Then
               IMIN = IDONT (ICRS)
               XMIN = XWRK1
            End If
         End Do
         IWRK1 = IMAX
         Do ICRS = NDON+1, IDEB+7
               IDONT (ICRS) = IWRK1
               If (IWRK1 == IMAX) Then
                  IWRK1 = IMIN
               Else
                  IWRK1 = IMAX
               End If
         End Do
!
         Do ICRS = IDEB+2, IDEB+7
            IWRK = IDONT (ICRS)
            XWRK = XDONT (IWRK)
            Do IDCR = ICRS - 1, IDEB+1, - 1
               If (XWRK >= XDONT (IDONT(IDCR))) Exit
               IDONT (IDCR+1) = IDONT (IDCR)
            End Do
            IDONT (IDCR+1) = IWRK
         End Do
         IDON1 = IDON1 + 1
         IWRKT (IDON1) = IDONT (IDEB + 3)
!
      End If
!
!  Find XMED7, the median of the medians
!
         IMED7 = i_med (IWRKT, IDON1)
         XMED7 = XDONT (IMED7)
!
!  Count how many values are not higher than (and how many equal to) XMED7
!  This number is at least 4 * 1/2 * (N/7) : 4 values in each of the
!  subsets where the median is lower than the median of medians. For similar
!  reasons, we also have at least 2N/7 values not lower than XMED7. At the
!  same time, we find in each subset the index of the last value < XMED7,
!  and that of the first > XMED7. These indices will be used to restrict the
!  search for the median as the Kth element in the subset (> or <) where
!  we know it to be.
!
         NMED = (NTRI+1) / 2
         IDON1 = 1
         NLEQ = 0
         NEQU = 0
         Do IDON = 1, NTRI, 7
            IMED = IDON+3
            If (XDONT (IDONT (IMED)) > XMED7) Then
                  IMED = IMED - 2
                  If (XDONT (IDONT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDONT (IDONT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Else If (XDONT (IDONT (IMED)) < XMED7) Then
                  IMED = IMED + 2
                  If (XDONT (IDONT (IMED)) > XMED7) Then
                     IMED = IMED - 1
                  Else If (XDONT (IDONT (IMED)) < XMED7) Then
                     IMED = IMED + 1
                  Endif
            Endif
            If (XDONT (IDONT (IMED)) > XMED7) Then
               NLEQ = NLEQ + IMED - IDON
               IENDT (IDON1) = IMED - 1
               ISTRT (IDON1) = IMED
            Else If (XDONT (IDONT (IMED)) < XMED7) Then
               NLEQ = NLEQ + IMED - IDON + 1
               IENDT (IDON1) = IMED
               ISTRT (IDON1) = IMED + 1
            Else                    !       If (XDONT (IDONT (IMED)) == XMED7)
               NLEQ = NLEQ + IMED - IDON + 1
               NEQU = NEQU + 1
               IENDT (IDON1) = IMED - 1
               Do IMED1 = IMED - 1, IDON, -1
                  If (XDONT (IDONT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     IENDT (IDON1) = IMED1 - 1
                  Else
                     Exit
                  End If
               End Do
               ISTRT (IDON1) = IMED + 1
               Do IMED1 = IMED + 1, IDON + 6
                  If (XDONT (IDONT (IMED1)) == XMED7) Then
                     NEQU = NEQU + 1
                     NLEQ = NLEQ + 1
                     ISTRT (IDON1) = IMED1 + 1
                  Else
                     Exit
                  End If
               End Do
            Endif
            IDON1 = IDON1 + 1
         End Do
!
!  Carry out a partial insertion sort to find the Kth smallest of the
!  large values, or the Kth largest of the small values, according to
!  what is needed.
!
         NMED = (NTRI+1) / 2
         If (NLEQ - NEQU + 1 <= NMED) Then
            If (NLEQ < NMED) Then   !      Not enough low values
                IWRK1 = IMAX
                XWRK1 = XDONT (IWRK1)
                NORD = NMED - NLEQ
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                IDCR = 0
                Do IDON = 1, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDONT (IDONT (ICRS)) < XWRK1) Then
                            IWRK = IDONT (ICRS)
                            XWRK = XDONT (IWRK)
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDONT (IDONT (IDCR))) Exit
                               IDONT  (IDCR+1) = IDONT (IDCR)
                            End Do
                            IDONT (IDCR+1) = IWRK
                            IWRK1 = IDONT (ICRS1)
                            XWRK1 = XDONT (IWRK1)
                         Else
                           If (ICRS2 < NORD) Then
                              IDONT (ICRS1) = IDONT (ICRS)
                              IWRK1 = IDONT (ICRS1)
                              XWRK1 = XDONT (IWRK1)
                           Endif
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                      If (ICRS2 >= NORD) Exit
                End Do
                Do IDON = IDON+7, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = ISTRT (IDON1), IDON + 6
                         If (XDONT (IDONT (ICRS)) >= XWRK1) Exit
                         IWRK = IDONT (ICRS)
                         XWRK = XDONT (IWRK)
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK >= XDONT (IDONT (IDCR))) Exit
                               IDONT  (IDCR+1) = IDONT (IDCR)
                         End Do
                         IDONT (IDCR+1) = IWRK
                         IWRK1 = IDONT (ICRS1)
                         XWRK1 = XDONT (IWRK1)
                      End Do
                End Do
                ires_med = IWRK1
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
            Else
                ires_med = IMED7
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
            End If
         Else                       !      If (NLEQ > NMED)
!                                          Not enough high values
                XWRK1 = -XHUGE
                NORD = NLEQ - NEQU - NMED + 1
                IDON1 = 0
                ICRS1 = 1
                ICRS2 = 0
                Do IDON = 1, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = IDON, IENDT (IDON1)
                         If (XDONT(IDONT (ICRS)) > XWRK1) Then
                            IWRK = IDONT (ICRS)
                            XWRK = XDONT (IWRK)
                            IDCR = ICRS1 - 1
                            Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDONT(IDONT (IDCR))) Exit
                               IDONT (IDCR+1) = IDONT (IDCR)
                            End Do
                            IDONT (IDCR+1) = IWRK
                            IWRK1 = IDONT(ICRS1)
                            XWRK1 = XDONT(IWRK1)
                         Else
                            If (ICRS2 < NORD) Then
                               IDONT (ICRS1) = IDONT (ICRS)
                               IWRK1 = IDONT(ICRS1)
                               XWRK1 = XDONT(IWRK1)
                            End If
                         End If
                         ICRS1 = MIN (NORD, ICRS1 + 1)
                         ICRS2 = MIN (NORD, ICRS2 + 1)
                      End Do
                      If (ICRS2 >= NORD) Exit
                End Do
                Do IDON = IDON+7, NTRI, 7
                      IDON1 = IDON1 + 1
                      Do ICRS = IENDT (IDON1), IDON, -1
                         If (XDONT(IDONT (ICRS)) <= XWRK1) Exit
                         IWRK = IDONT (ICRS)
                         XWRK = XDONT (IWRK)
                         IDCR = ICRS1 - 1
                         Do IDCR = ICRS1 - 1, 1, - 1
                               If (XWRK <= XDONT(IDONT (IDCR))) Exit
                               IDONT (IDCR+1) = IDONT (IDCR)
                         End Do
                         IDONT (IDCR+1) = IWRK
                         IWRK1 = IDONT(ICRS1)
                         XWRK1 = XDONT(IWRK1)
                      End Do
                End Do
!
                ires_med = IWRK1
                Deallocate (ISTRT, IENDT, IWRKT)
                Return
         End If
   END Function i_med
End Function INDMED
